"""Bernardi-Raugel elements on simplices.

This element's definition appears in https://doi.org/10.2307/2007793
(Bernardi and Raugel, 1985)
"""

import typing

import sympy

from symfem.elements.lagrange import Lagrange
from symfem.finite_element import CiarletElement
from symfem.functionals import (
    DivergenceIntegralMoment,
    DotPointEvaluation,
    ListOfFunctionals,
    NormalIntegralMoment,
)
from symfem.functions import FunctionInput
from symfem.moments import make_integral_moment_dofs
from symfem.polynomials import polynomial_set_vector
from symfem.references import NonDefaultReferenceError, Reference
from symfem.symbols import x

__all__ = ["BernardiRaugel"]


class BernardiRaugel(CiarletElement):
    """Bernardi-Raugel Hdiv finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        tdim = reference.tdim

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order)

        p = Lagrange(reference, 1, variant="equispaced")

        for i in range(reference.sub_entity_count(reference.tdim - 1)):
            sub_e = reference.sub_entity(reference.tdim - 1, i)
            bubble = sympy.Integer(1)
            for j in reference.sub_entities(reference.tdim - 1)[i]:
                bubble *= p.get_basis_function(j)
            poly.append(tuple(bubble * j for j in sub_e.normal()))

        dofs: ListOfFunctionals = []

        # Evaluation at vertices
        for n in range(reference.sub_entity_count(0)):
            vertex = reference.sub_entity(0, n)
            v = vertex.vertices[0]
            for i in range(tdim):
                direction = tuple(1 if i == j else 0 for j in range(tdim))
                dofs.append(
                    DotPointEvaluation(
                        reference,
                        v,
                        direction,
                        entity=(0, n),
                        mapping="identity",
                    )
                )

        dofs += make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, Lagrange, 0, "contravariant", {"variant": "equispaced"}),
        )

        if order > 1:
            assert order == 2 and reference.name == "tetrahedron"

            for i in range(reference.tdim):
                bf = p.get_basis_functions()
                poly.append(
                    tuple(
                        bf[0] * bf[1] * bf[2] * bf[3] if j == i else 0
                        for j in range(reference.tdim)
                    )
                )

            for e_n, edge in enumerate(reference.edges):
                v1 = reference.vertices[edge[0]]
                v2 = reference.vertices[edge[1]]
                midpoint = tuple(sympy.Rational(i + j, 2) for i, j in zip(v1, v2))
                for i in range(tdim):
                    direction = tuple(1 if i == j else 0 for j in range(tdim))
                    dofs.append(
                        DotPointEvaluation(
                            reference, midpoint, direction, entity=(1, e_n), mapping="identity"
                        )
                    )

            p = Lagrange(reference, 0, variant="equispaced")
            for i in range(3):
                dofs.append(
                    DivergenceIntegralMoment(
                        reference, x[i], p.dofs[0], entity=(3, 0), mapping="identity"
                    )
                )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + self.reference.tdim - 1

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order + self.reference.tdim - 1

    names = ["Bernardi-Raugel"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    max_order = {"triangle": 1, "tetrahedron": 2}
    continuity = "L2"
    value_type = "vector"
    last_updated = "2024.10.1"



"""Morley-Wang-Xu elements on simplices.

This element's definition appears in https://doi.org/10.1090/S0025-5718-2012-02611-1
(Wang, Xu, 2013)
"""

import typing

from symfem.finite_element import CiarletElement
from symfem.functionals import (
    IntegralAgainst,
    IntegralOfDirectionalMultiderivative,
    ListOfFunctionals,
    PointEvaluation,
)
from symfem.functions import FunctionInput
from symfem.polynomials import polynomial_set_1d
from symfem.references import NonDefaultReferenceError, Reference

__all__ = ["MorleyWangXu"]


class MorleyWangXu(CiarletElement):
    """Morley-Wang-Xu finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert order <= reference.tdim
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)

        dofs: ListOfFunctionals = []

        if order == 1:
            if reference.tdim == 1:
                for v_n, v in enumerate(reference.vertices):
                    dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            else:
                dim = reference.tdim - 1
                for facet_n in range(reference.sub_entity_count(dim)):
                    facet = reference.sub_entity(dim, facet_n)
                    dofs.append(
                        IntegralAgainst(reference, 1 / facet.jacobian(), entity=(dim, facet_n))
                    )
        elif order == 2:
            if reference.tdim == 2:
                for v_n, v in enumerate(reference.vertices):
                    dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            else:
                dim = reference.tdim - 2
                for ridge_n in range(reference.sub_entity_count(dim)):
                    ridge = reference.sub_entity(dim, ridge_n)
                    dofs.append(
                        IntegralAgainst(reference, 1 / ridge.jacobian(), entity=(dim, ridge_n))
                    )
            dim = reference.tdim - 1
            for facet_n in range(reference.sub_entity_count(dim)):
                facet = reference.sub_entity(dim, facet_n)
                dofs.append(
                    IntegralOfDirectionalMultiderivative(
                        reference,
                        (facet.normal(),),
                        (1,),
                        (dim, facet_n),
                        scale=1 / facet.jacobian(),
                    )
                )
        else:
            assert order == reference.tdim == 3
            for v_n, v in enumerate(reference.vertices):
                dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            for e_n, vs in enumerate(reference.sub_entities(1)):
                subentity = reference.sub_entity(1, e_n)
                volume = subentity.jacobian()
                normals = []
                for f_n, f_vs in enumerate(reference.sub_entities(2)):
                    if vs[0] in f_vs and vs[1] in f_vs:
                        face = reference.sub_entity(2, f_n)
                        normals.append(face.normal())
                for orders in [(1, 0), (0, 1)]:
                    dofs.append(
                        IntegralOfDirectionalMultiderivative(
                            reference, tuple(normals), orders, (1, e_n), scale=1 / volume
                        )
                    )
            for f_n, vs in enumerate(reference.sub_entities(2)):
                subentity = reference.sub_entity(2, f_n)
                volume = subentity.jacobian()
                dofs.append(
                    IntegralOfDirectionalMultiderivative(
                        reference, (subentity.normal(),), (2,), (2, f_n), scale=1 / volume
                    )
                )

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order

    names = ["Morley-Wang-Xu", "MWX"]
    references = ["interval", "triangle", "tetrahedron"]
    min_order = 1
    max_order = {"interval": 1, "triangle": 2, "tetrahedron": 3}
    # continuity = "C{order}"
    continuity = "C0"
    value_type = "scalar"
    last_updated = "2023.06"



"""Bell elements on triangle.

This element's definition is given in https://doi.org/10.1002/nme.1620010108 (Bell, 1969)
"""

import typing

from symfem.finite_element import CiarletElement
from symfem.functionals import DerivativePointEvaluation, ListOfFunctionals, PointEvaluation
from symfem.functions import FunctionInput
from symfem.polynomials import polynomial_set_1d
from symfem.references import Reference
from symfem.symbols import x

__all__ = ["Bell"]


class Bell(CiarletElement):
    """Bell finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        assert reference.name == "triangle"
        dofs: ListOfFunctionals = []
        for v_n, v in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, v, (1, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, v, (0, 1), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, v, (2, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, v, (1, 1), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, v, (0, 2), entity=(0, v_n)))
        self.variant = variant

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, 4)
        poly.append(x[0] ** 5 - x[1] ** 5)
        poly.append(x[0] ** 3 * x[1] ** 2 - x[0] ** 2 * x[1] ** 3)
        poly.append(5 * x[0] ** 2 * x[1] ** 3 - x[0] ** 5)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return 4

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return 5

    @property
    def polynomial_subdegree(self) -> int:
        return 4

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return 5

    names = ["Bell"]
    references = ["triangle"]
    min_order = 4
    max_order = 4
    continuity = "C1"
    value_type = "scalar"
    last_updated = "2025.03"
    _max_continuity_test_order = 3



"""Nedelec elements on prisms."""

import typing

from symfem.elements.lagrange import Lagrange, VectorLagrange
from symfem.elements.q import RaviartThomas as QRT
from symfem.finite_element import CiarletElement
from symfem.functionals import (
    IntegralAgainst,
    IntegralMoment,
    ListOfFunctionals,
    TangentIntegralMoment,
)
from symfem.functions import FunctionInput
from symfem.moments import make_integral_moment_dofs
from symfem.polynomials import Hcurl_polynomials, polynomial_set_1d, polynomial_set_vector
from symfem.references import NonDefaultReferenceError, Reference
from symfem.symbols import x

__all__ = ["Nedelec"]


class Nedelec(CiarletElement):
    """Nedelec Hcurl finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        from symfem import create_reference

        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        poly: typing.List[FunctionInput] = []
        poly += [
            (i[0] * j, i[1] * j, 0)
            for i in polynomial_set_vector(2, 2, order) + Hcurl_polynomials(2, 2, order + 1)
            for j in polynomial_set_1d(1, order + 1, x[2:])
        ]
        poly += [
            (0, 0, i * j)
            for i in polynomial_set_1d(2, order + 1, x[:2])
            for j in polynomial_set_1d(1, order, x[2:])
        ]

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, Lagrange, order, {"variant": variant}),
            faces={
                "triangle": (
                    IntegralMoment,
                    VectorLagrange,
                    order - 1,
                    "covariant",
                    {"variant": variant},
                ),
                "quadrilateral": (
                    IntegralMoment,
                    QRT,
                    order - 1,
                    "covariant",
                    {"variant": variant},
                ),
            },
        )

        triangle = create_reference("triangle")
        interval = create_reference("interval")

        if order >= 1:
            space1 = VectorLagrange(triangle, order - 1, variant)
            space2 = Lagrange(interval, order - 1, variant)

            if order > 1:
                raise NotImplementedError()
            # TODO: correct these for order > 1
            for i in range(space1.space_dim):
                for j in range(space2.space_dim):
                    f = (
                        space2.get_basis_function(j) * space1.get_basis_function(i)[0],
                        space2.get_basis_function(j) * space1.get_basis_function(i)[1],
                        0,
                    )
                    dofs.append(IntegralAgainst(reference, f, entity=(3, 0), mapping="covariant"))

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return (self.order + 1) * 2

    names = ["Nedelec", "Ncurl"]
    references = ["prism"]
    min_order = 0
    max_order = 1
    continuity = "H(curl)"
    value_type = "vector"
    last_updated = "2025.03"



"""TiNiest Tensor product (TNT) elements.

These elements' definitions appear in https://doi.org/10.1090/S0025-5718-2013-02729-9
(Cockburn, Qiu, 2013)
"""

import typing
from itertools import product

import sympy

from symfem.elements.q import Q
from symfem.finite_element import CiarletElement
from symfem.functionals import (
    DerivativeIntegralMoment,
    IntegralAgainst,
    ListOfFunctionals,
    NormalIntegralMoment,
    PointEvaluation,
    TangentIntegralMoment,
)
from symfem.functions import FunctionInput, ScalarFunction, VectorFunction
from symfem.moments import make_integral_moment_dofs
from symfem.polynomials import orthogonal_basis, quolynomial_set_1d, quolynomial_set_vector
from symfem.references import NonDefaultReferenceError, Reference
from symfem.symbols import t, x

__all__ = ["p", "b", "TNT", "TNTcurl", "TNTdiv"]


def p(k: int, v: sympy.core.symbol.Symbol) -> ScalarFunction:
    """Return the kth Legendre polynomial.

    Args:
        k: k
        v: The variable to use

    Returns:
        The kth Legendre polynomial
    """
    return orthogonal_basis("interval", k, 0, [v])[0][-1]


def b(k: int, v: sympy.core.symbol.Symbol) -> ScalarFunction:
    """Return the function B_k.

    This function is defined on page 4 (606) of
    https://doi.org/10.1090/S0025-5718-2013-02729-9 (Cockburn, Qiu, 2013).

    Args:
        k: k
        v: The variable to use

    Returns:
        The function B_k
    """
    if k == 1:
        return ScalarFunction(0)
    return (p(k, v) - p(k - 2, v)) / (4 + k - 2)


class TNT(CiarletElement):
    """TiNiest Tensor scalar finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        poly: typing.List[FunctionInput] = []
        poly += quolynomial_set_1d(reference.tdim, order - 1)
        if reference.tdim == 2:
            for i in range(2):
                variables = [x[j] for j in range(2) if j != i]
                for f in [1 - variables[0], variables[0]]:
                    poly.append(f * b(order, x[i]))

        elif reference.tdim == 3:
            for i in range(3):
                variables = [x[j] for j in range(3) if j != i]
                for f0 in [1 - variables[0], variables[0]]:
                    for f1 in [1 - variables[1], variables[1]]:
                        poly.append(f0 * f1 * b(order, x[i]))

        dofs: ListOfFunctionals = []
        for i, v in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, v, entity=(0, i)))

        for i in range(1, order):
            f = i * t[0] ** (i - 1)
            for edge_n in range(reference.sub_entity_count(1)):
                dofs.append(IntegralAgainst(reference, f, entity=(1, edge_n), mapping="identity"))

        for i in range(1, order - 1):
            for j in range(1, order - 1):
                f = t[0] ** i * (t[0] - 1) * t[1] ** j * (t[1] - 1)
                delta_f = (f.diff(t[0]).diff(t[0]) + f.diff(t[1]).diff(t[1])).expand()
                for face_n in range(reference.sub_entity_count(2)):
                    dofs.append(
                        IntegralAgainst(reference, delta_f, entity=(2, face_n), mapping="identity")
                    )

        if reference.tdim == 3:
            dummy_dof = PointEvaluation(reference, reference.midpoint(), (3, 0))
            for ii in product(range(1, order - 1), repeat=3):
                f = sympy.Integer(1)
                for j, k in zip(ii, x):
                    f *= k**j * (k - 1)
                grad_f = tuple(sympy.S(j).expand() for j in ScalarFunction(f).grad(3))
                dofs.append(
                    DerivativeIntegralMoment(
                        reference, 1, grad_f, dummy_dof, entity=(3, 0), mapping="identity"
                    )
                )

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return self.order - 1

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return max((self.order - 1) * self.reference.tdim, (self.order - 1) + self.reference.tdim)

    names = ["tiniest tensor", "TNT"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 2
    continuity = "C0"
    value_type = "scalar"
    last_updated = "2025.03"


class TNTcurl(CiarletElement):
    """TiNiest Tensor Hcurl finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        poly: typing.List[FunctionInput] = []
        poly += quolynomial_set_vector(reference.tdim, reference.tdim, order)
        if reference.tdim == 2:
            for ii in product([0, 1], repeat=2):
                if sum(ii) != 0:
                    poly.append(
                        tuple(
                            sympy.S(j).expand()
                            for j in [
                                p(order, ii[0] * x[0]) * b(order + 1, ii[1] * x[1]),
                                -b(order + 1, ii[0] * x[0]) * p(order, ii[1] * x[1]),
                            ]
                        )
                    )
        else:
            face_poly = []
            for ii in product([0, 1], repeat=2):
                if sum(ii) != 0:
                    face_poly.append(
                        tuple(
                            sympy.S(j).expand()
                            for j in [
                                b(order + 1, ii[0] * t[0]) * p(order, ii[1] * t[1]),
                                p(order, ii[0] * t[0]) * b(order + 1, ii[1] * t[1]),
                            ]
                        )
                    )
            for lamb_n in [
                (x[0], 0, 0),
                (1 - x[0], 0, 0),
                (0, x[1], 0),
                (0, 1 - x[1], 0),
                (0, 0, x[2]),
                (0, 0, 1 - x[2]),
            ]:
                variables = tuple(i for i, j in enumerate(lamb_n) if j == 0)
                for pf in face_poly:
                    psub = VectorFunction(pf).subs(t[:2], [x[j] for j in variables])
                    pc = VectorFunction(lamb_n).cross(
                        VectorFunction(
                            [psub[variables.index(i)] if i in variables else 0 for i in range(3)]
                        )
                    )
                    poly.append(pc)

        dofs: ListOfFunctionals = []
        dofs += make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, Q, order, {"variant": variant}),
        )

        # Face moments
        face_moments = []
        for ii in product(range(order + 1), repeat=2):
            if sum(ii) > 0:
                f = x[0] ** ii[0] * x[1] ** ii[1]
                grad_f = ScalarFunction(f).grad(2)
                grad_f2 = VectorFunction((grad_f[1], -grad_f[0])).subs(x[:2], tuple(t[:2]))
                face_moments.append(grad_f2)

        for i in range(2, order + 1):
            for j in range(2, order + 1):
                face_moments.append(
                    VectorFunction(
                        (
                            t[1] ** (j - 1) * (1 - t[1]) * t[0] ** (i - 2) * (i * t[0] - i + 1),
                            -(t[0] ** (i - 1)) * (1 - t[0]) * t[1] ** (j - 2) * (j - 1 - j * t[1]),
                        )
                    )
                )
        if reference.tdim == 2:
            for f in face_moments:
                dofs.append(IntegralAgainst(reference, f, entity=(2, 0), mapping="contravariant"))
        elif reference.tdim == 3:
            for face_n in range(6):
                for f in face_moments:
                    dofs.append(
                        IntegralAgainst(reference, f, entity=(2, face_n), mapping="contravariant")
                    )

        # Interior Moments
        if reference.tdim == 3:
            for i in range(1, order):
                for j in range(1, order):
                    for k in range(order + 1):
                        f = (x[0] ** k * x[1] ** i * (1 - x[1]) * x[2] ** j * (1 - x[2]), 0, 0)
                        dofs.append(
                            IntegralAgainst(
                                reference,
                                VectorFunction(f).curl().curl(),
                                entity=(3, 0),
                                mapping="covariant",
                            )
                        )

                        f = (0, x[1] ** k * x[0] ** i * (1 - x[0]) * x[2] ** j * (1 - x[2]), 0)
                        dofs.append(
                            IntegralAgainst(
                                reference,
                                VectorFunction(f).curl().curl(),
                                entity=(3, 0),
                                mapping="covariant",
                            )
                        )

                        if k in [0, 2]:
                            f = (0, 0, x[2] ** k * x[0] ** i * (1 - x[0]) * x[1] ** j * (1 - x[1]))
                            dofs.append(
                                IntegralAgainst(
                                    reference,
                                    VectorFunction(f).curl().curl(),
                                    entity=(3, 0),
                                    mapping="covariant",
                                )
                            )

            for i in range(2, order + 1):
                for j in range(2, order + 1):
                    for k in range(2, order + 1):
                        f = x[0] ** (i - 1) * x[0] ** i
                        f *= x[1] ** (j - 1) * x[1] ** j
                        f *= x[2] ** (k - 1) * x[2] ** k
                        grad_f = ScalarFunction(f).grad(3)
                        dofs.append(
                            IntegralAgainst(
                                reference, grad_f, entity=(3, 0), mapping="contravariant"
                            )
                        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order * self.reference.tdim + 1

    names = ["tiniest tensor Hcurl", "TNTcurl"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    continuity = "H(curl)"
    value_type = "vector"
    last_updated = "2023.06"


class TNTdiv(CiarletElement):
    """TiNiest Tensor Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        poly: typing.List[FunctionInput] = []
        poly += quolynomial_set_vector(reference.tdim, reference.tdim, order)
        if reference.tdim == 2:
            for ii in product([0, 1], repeat=2):
                if sum(ii) != 0:
                    poly.append(
                        tuple(
                            sympy.S(j).expand()
                            for j in [
                                b(order + 1, ii[0] * x[0]) * p(order, ii[1] * x[1]),
                                p(order, ii[0] * x[0]) * b(order + 1, ii[1] * x[1]),
                            ]
                        )
                    )
        else:
            for ii in product([0, 1], repeat=3):
                if sum(ii) != 0:
                    poly.append(
                        (
                            b(order + 1, ii[0] * x[0])
                            * p(order, ii[1] * x[1])
                            * p(order, ii[2] * x[2]),
                            p(order, ii[0] * x[0])
                            * b(order + 1, ii[1] * x[1])
                            * p(order, ii[2] * x[2]),
                            p(order, ii[0] * x[0])
                            * p(order, ii[1] * x[1])
                            * b(order + 1, ii[2] * x[2]),
                        )
                    )

        dofs: ListOfFunctionals = []
        dofs += make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, Q, order, {"variant": variant}),
        )

        for ii in product(range(order + 1), repeat=reference.tdim):
            if sum(ii) > 0:
                if reference.tdim == 2:
                    f = x[0] ** ii[0] * x[1] ** ii[1]
                else:
                    f = x[0] ** ii[0] * x[1] ** ii[1] * x[2] ** ii[2]
                grad_f = ScalarFunction(f).grad(reference.tdim)
                dofs.append(
                    IntegralAgainst(
                        reference, grad_f, entity=(reference.tdim, 0), mapping="covariant"
                    )
                )

        if reference.tdim == 2:
            for i in range(2, order + 1):
                for j in range(2, order + 1):
                    f = (
                        x[0] ** (i - 1) * (1 - x[0]) * x[1] ** (j - 2) * (j - 1 - j * x[1]),
                        x[1] ** (j - 1) * (1 - x[1]) * x[0] ** (i - 2) * (i * x[0] - i + 1),
                    )
                    dofs.append(
                        IntegralAgainst(
                            reference, f, entity=(reference.tdim, 0), mapping="covariant"
                        )
                    )
        if reference.tdim == 3:
            for i in range(2, order + 1):
                for j in range(2, order + 1):
                    for k in range(order + 1):
                        f = (
                            x[2] ** k
                            * x[0] ** (i - 1)
                            * (1 - x[0])
                            * x[2] ** (j - 2)
                            * (j - 1 - j * x[1]),
                            x[2] ** k
                            * x[1] ** (j - 1)
                            * (1 - x[1])
                            * x[0] ** (i - 2)
                            * (i * x[0] - i + 1),
                            0,
                        )
                        dofs.append(
                            IntegralAgainst(
                                reference, f, entity=(reference.tdim, 0), mapping="covariant"
                            )
                        )
                        f = (
                            x[1] ** k
                            * x[0] ** (i - 1)
                            * (1 - x[0])
                            * x[2] ** (j - 2)
                            * (j - 1 - j * x[2]),
                            0,
                            x[1] ** k
                            * x[2] ** (j - 1)
                            * (1 - x[2])
                            * x[0] ** (i - 2)
                            * (i * x[0] - i + 1),
                        )
                        dofs.append(
                            IntegralAgainst(
                                reference, f, entity=(reference.tdim, 0), mapping="covariant"
                            )
                        )
                        if k in [0, 2]:
                            f = (
                                0,
                                x[0] ** k
                                * x[1] ** (i - 1)
                                * (1 - x[1])
                                * x[2] ** (j - 2)
                                * (j - 1 - j * x[2]),
                                x[0] ** k
                                * x[2] ** (j - 1)
                                * (1 - x[2])
                                * x[1] ** (i - 2)
                                * (i * x[1] - i + 1),
                            )
                            dofs.append(
                                IntegralAgainst(
                                    reference, f, entity=(reference.tdim, 0), mapping="covariant"
                                )
                            )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order * self.reference.tdim + 1

    names = ["tiniest tensor Hdiv", "TNTdiv"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    continuity = "H(div)"
    value_type = "vector"
    last_updated = "2023.06"



"""Lagrange elements on a pyramid.

This element's definition appears in https://doi.org/10.1007/s10915-009-9334-9
(Bergot, Cohen, Durufle, 2010)
"""

import typing
from itertools import product

import sympy

from symfem.finite_element import CiarletElement
from symfem.functionals import IntegralAgainst, ListOfFunctionals, PointEvaluation
from symfem.functions import FunctionInput
from symfem.polynomials import orthonormal_basis, pyramid_polynomial_set_1d
from symfem.quadrature import get_quadrature
from symfem.references import NonDefaultReferenceError, Reference

__all__ = ["Lagrange"]


class Lagrange(CiarletElement):
    """Lagrange finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        dofs: ListOfFunctionals = []
        if variant == "legendre":
            basis = orthonormal_basis(reference.name, order, 0)[0]
            for f in basis:
                dofs.append(IntegralAgainst(reference, f, (reference.tdim, 0)))
        elif order == 0:
            dofs = [
                PointEvaluation(
                    reference,
                    tuple(sympy.Rational(1, reference.tdim + 1) for i in range(reference.tdim)),
                    entity=(reference.tdim, 0),
                )
            ]
        elif variant == "lobatto":
            raise NotImplementedError()
        else:
            if variant == "gl":
                points, _ = get_quadrature("legendre", order + 1)
            else:
                points, _ = get_quadrature(variant, order + 1)

            # Vertices
            for v_n, v in enumerate(reference.vertices):
                dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            # Edges
            for e_n in range(reference.sub_entity_count(1)):
                entity = reference.sub_entity(1, e_n)
                for i in range(1, order):
                    dofs.append(
                        PointEvaluation(
                            reference,
                            tuple(
                                o + entity.axes[0][j] * points[i]
                                for j, o in enumerate(entity.origin)
                            ),
                            entity=(1, e_n),
                        )
                    )
            # Faces
            for e_n in range(reference.sub_entity_count(2)):
                entity = reference.sub_entity(2, e_n)
                for ii in product(range(1, order), repeat=2):
                    if len(entity.vertices) == 4 or sum(ii) < order:
                        dofs.append(
                            PointEvaluation(
                                reference,
                                tuple(
                                    o + sum(a[j] * points[b] for a, b in zip(entity.axes, ii[::-1]))
                                    for j, o in enumerate(entity.origin)
                                ),
                                entity=(2, e_n),
                            )
                        )

            # Interior
            for ii in product(range(1, order), repeat=3):
                if max(ii[0], ii[1]) + ii[2] < order:
                    dofs.append(
                        PointEvaluation(
                            reference,
                            tuple(
                                o + sum(a[j] * points[b] for a, b in zip(reference.axes, ii))
                                for j, o in enumerate(reference.origin)
                            ),
                            entity=(3, 0),
                        )
                    )

        poly: typing.List[FunctionInput] = []
        poly += pyramid_polynomial_set_1d(reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return None

    names = ["Lagrange", "P"]
    references = ["pyramid"]
    min_order = 0
    continuity = "C0"
    value_type = "scalar"
    last_updated = "2024.09"



"""Taylor element on an interval, triangle or tetrahedron."""

import typing
from itertools import product

from symfem.elements.lagrange import Lagrange
from symfem.finite_element import CiarletElement
from symfem.functionals import DerivativePointEvaluation, IntegralMoment, ListOfFunctionals
from symfem.functions import FunctionInput
from symfem.moments import make_integral_moment_dofs
from symfem.polynomials import polynomial_set_1d
from symfem.references import Reference

__all__ = ["Taylor"]


class Taylor(CiarletElement):
    """Taylor finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            cells=(IntegralMoment, Lagrange, 0, {"variant": "equispaced"}),
        )
        for i in product(range(order + 1), repeat=reference.tdim):
            if 1 <= sum(i) <= order:
                dofs.append(
                    DerivativePointEvaluation(
                        reference, reference.midpoint(), i, entity=(reference.tdim, 0)
                    )
                )

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order

    names = ["Taylor", "discontinuous Taylor"]
    references = ["interval", "triangle", "tetrahedron"]
    min_order = 0
    continuity = "L2"
    value_type = "scalar"
    last_updated = "2023.06"



"""Brezzi-Douglas-Fortin-Marini elements.

This element's definition appears in https://doi.org/10.1051/m2an/1987210405811
(Brezzi, Douglas, Fortin, Marini, 1987) and
https://doi.org/10.1007/978-1-4612-3172-1 (Brezzi, Fortin, 1991)
"""

import typing

from symfem.elements.dpc import DPC, VectorDPC
from symfem.elements.lagrange import Lagrange
from symfem.elements.nedelec import NedelecFirstKind
from symfem.finite_element import CiarletElement
from symfem.functionals import IntegralMoment, ListOfFunctionals, NormalIntegralMoment
from symfem.functions import FunctionInput
from symfem.moments import make_integral_moment_dofs
from symfem.polynomials import polynomial_set_vector
from symfem.references import NonDefaultReferenceError, Reference
from symfem.symbols import x

__all__ = ["bdfm_polyset", "BDFM"]


def bdfm_polyset(reference: Reference, order: int) -> typing.List[FunctionInput]:
    """Create the polynomial basis for a BDFM element.

    Args:
        reference: The reference cell
        order: The polynomial order

    Returns:
        The polynomial basis
    """
    dim = reference.tdim
    pset: typing.List[FunctionInput] = []
    pset += polynomial_set_vector(dim, dim, order)
    if reference.name == "quadrilateral":
        for i in range(1, order + 2):
            j = order + 1 - i
            pset.append((x[0] ** i * x[1] ** j, 0))
            pset.append((0, x[1] ** i * x[0] ** j))
    elif reference.name == "triangle":
        for i in range(order):
            p = x[0] ** i * x[1] ** (order - 1 - i)
            pset.append((x[0] * (x[0] + x[1]) * p, 0))
            pset.append((0, x[1] * (x[0] + x[1]) * p))
        p = x[0] ** order
        pset.append((x[0] * p, x[1] * p))
    elif reference.name == "hexahedron":
        for i in range(1, order + 2):
            for j in range(order + 2 - i):
                k = order + 1 - i - j
                pset.append((x[0] ** i * x[1] ** j * x[2] ** k, 0, 0))
                pset.append((0, x[1] ** i * x[0] ** j * x[2] ** k, 0))
                pset.append((0, 0, x[2] ** i * x[0] ** j * x[1] ** k))
    elif reference.name == "tetrahedron":
        for i in range(order):
            for j in range(order - i):
                p = x[0] ** i * x[1] ** j * x[2] ** (order - 1 - i - j)
                pset.append((x[0] * (x[0] + x[1] + x[2]) * p, 0, 0))
                pset.append((0, x[1] * (x[0] + x[1] + x[2]) * p, 0))
                pset.append((0, 0, x[2] * (x[0] + x[1] + x[2]) * p))
        for i in range(order + 1):
            p = x[0] ** i * x[1] ** (order - i)
            pset.append((x[0] * p, x[1] * p, x[2] * p))
        for i in range(1, order + 1):
            p = x[0] ** i * x[1] ** (order - i)
            pset.append((p * (x[0] + x[1]), 0, x[1] * p))

    return pset


class BDFM(CiarletElement):
    """Brezzi-Douglas-Fortin-Marini Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        poly = bdfm_polyset(reference, order)

        dofs: ListOfFunctionals = []
        if reference.name in ["triangle", "tetrahedron"]:
            dofs = make_integral_moment_dofs(
                reference,
                facets=(NormalIntegralMoment, Lagrange, order, {"variant": variant}),
                cells=(IntegralMoment, NedelecFirstKind, order - 1, {"variant": variant}),
            )
        else:
            dofs = make_integral_moment_dofs(
                reference,
                facets=(NormalIntegralMoment, DPC, order, {"variant": variant}),
                cells=(IntegralMoment, VectorDPC, order - 1, {"variant": variant}),
            )
        self.variant = variant

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        if self.reference.name in ["triangle", "tetrahedron"]:
            return self.order
        else:
            return (self.order + 1) // self.reference.tdim

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    names = ["Brezzi-Douglas-Fortin-Marini", "BDFM"]
    references = ["triangle", "quadrilateral", "hexahedron", "tetrahedron"]
    min_order = 0
    continuity = "H(div)"
    value_type = "vector"
    last_updated = "2025.03"



"""Functions to cache matrices."""

import os
import typing
from hashlib import sha256

import sympy
from appdirs import user_cache_dir

if not os.path.isdir(user_cache_dir()):
    os.mkdir(user_cache_dir())
CACHE_DIR = user_cache_dir("symfem")
CACHE_FORMAT = "1"
if os.path.isfile(CACHE_DIR):
    os.remove(CACHE_DIR)
if not os.path.isdir(CACHE_DIR):
    os.mkdir(CACHE_DIR)

assert os.path.isdir(CACHE_DIR)

__all__ = ["load_cached_matrix", "save_cached_matrix", "matrix_to_string", "matrix_from_string"]


def load_cached_matrix(
    matrix_type: str, cache_id: str, size: typing.Tuple[int, int]
) -> typing.Union[sympy.matrices.dense.MutableDenseMatrix, None]:
    """Load a cached matrix.

    Args:
        matrix_type: The type of the matrix. This will be included in the filename.
        cache_id: The unique identifier of the matrix within this type

    Returns:
        The matrix
    """
    hashed_id = sha256(cache_id.encode("utf-8")).hexdigest()
    filename = os.path.join(CACHE_DIR, f"{matrix_type}{CACHE_FORMAT}-{hashed_id}.matrix")
    try:
        with open(filename) as f:
            mat = matrix_from_string(f.read())
        if mat.rows != size[0] or mat.cols != size[1]:
            return None
        return mat
    except FileNotFoundError:
        return None


def save_cached_matrix(
    matrix_type: str, cache_id: str, matrix: sympy.matrices.dense.MutableDenseMatrix
):
    """Save a matrix to the cache.

    Args:
        matrix_type: The type of the matrix. This will be included in the filename.
        cache_id: The unique identifier of the matrix within this type
        matrix: The matrix
    """
    hashed_id = sha256(cache_id.encode("utf-8")).hexdigest()
    filename = os.path.join(CACHE_DIR, f"{matrix_type}{CACHE_FORMAT}-{hashed_id}.matrix")
    with open(filename, "w") as f:
        f.write(matrix_to_string(matrix))


def matrix_to_string(m: sympy.matrices.dense.MutableDenseMatrix) -> str:
    """Convert a matrix to a string.

    Args:
        m: The matrix

    Returns:
        A representation of the matrix as a string
    """
    return ";".join(",".join(f"{m[i, j]!r}" for j in range(m.cols)) for i in range(m.rows))


def matrix_from_string(mstr: str) -> sympy.matrices.dense.MutableDenseMatrix:
    """Convert a string to a matrix.

    Args:
        mstr: The string in the format output by `matrix_to_string`

    Returns:
        The matrix
    """
    return sympy.Matrix([[sympy.S(j) for j in i.split(",")] for i in mstr.split(";")])



"""Transition elements on simplices."""

import typing
from itertools import product

import sympy

from symfem.elements.lagrange import Lagrange
from symfem.finite_element import CiarletElement
from symfem.functionals import ListOfFunctionals, PointEvaluation
from symfem.functions import FunctionInput, ScalarFunction
from symfem.polynomials import polynomial_set_1d
from symfem.quadrature import get_quadrature
from symfem.references import Reference
from symfem.symbols import x

__all__ = ["Transition"]


class Transition(CiarletElement):
    """Transition finite element."""

    def __init__(
        self,
        reference: Reference,
        order: int,
        edge_orders: typing.Optional[typing.List[int]] = None,
        face_orders: typing.Optional[typing.List[int]] = None,
        variant: str = "equispaced",
    ):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            edge_orders: the polynomial order for each edge
            face_orders: the polynomial order for each face
            variant: The variant of the element
        """
        if reference.name == "triangle":
            assert face_orders is None
            assert edge_orders is not None
            assert len(edge_orders) == 3
        elif reference.name == "tetrahedron":
            assert face_orders is not None
            assert edge_orders is not None
            assert len(face_orders) == 4
            assert len(edge_orders) == 6
        bubble_space = Lagrange(reference.default_reference(), 1)

        dofs: ListOfFunctionals = []
        for v_n, v in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, 1)

        for edim in range(1, 4):
            for e_n in range(reference.sub_entity_count(edim)):
                entity = reference.sub_entity(edim, e_n)
                ref_entity = reference.default_reference().sub_entity(edim, e_n)
                if edim == reference.tdim:
                    entity_order = order
                elif edim == 1:
                    assert edge_orders is not None
                    entity_order = edge_orders[e_n]
                elif edim == 2:
                    assert face_orders is not None
                    entity_order = face_orders[e_n]
                else:
                    raise RuntimeError("Could not find order for this entity.")

                # DOFs
                points, _ = get_quadrature(variant, entity_order + 1)
                for ii in product(range(1, entity_order), repeat=edim):
                    if sum(ii) < entity_order:
                        pt = entity.get_point([points[j] for j in ii])
                        dofs.append(PointEvaluation(reference, pt, entity=(edim, e_n)))

                # Basis
                if entity_order > edim:
                    if edim == reference.tdim:
                        bubble = sympy.Integer(1)
                        for f in bubble_space.get_basis_functions():
                            bubble *= f
                    elif edim == reference.tdim - 1:
                        bubble = sympy.Integer(1)
                        for i, f in enumerate(bubble_space.get_basis_functions()):
                            if i != e_n:
                                bubble *= f
                    else:
                        assert edim == 1 and reference.tdim == 3
                        bubble = sympy.Integer(1)
                        for i, f in enumerate(bubble_space.get_basis_functions()):
                            if i in reference.edges[e_n]:
                                bubble *= f
                    space = Lagrange(
                        entity.default_reference(), entity_order - edim - 1, variant=variant
                    )
                    variables = []
                    origin = ref_entity.vertices[0]
                    used = []
                    for p in ref_entity.vertices[1:]:
                        i = 0
                        while p[i] == origin[i] or origin[i] == 1 or i in used:
                            i += 1
                        used.append(i)
                        variables.append(origin[i] + (p[i] - origin[i]) * x[i])
                    for f in space.get_basis_functions():
                        assert isinstance(f, ScalarFunction)
                        poly.append(f.subs(x, variables) * bubble)

        poly = reference.map_polyset_from_default(poly)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)
        self.variant = variant
        self.face_orders = face_orders
        self.edge_orders = edge_orders

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {
            "variant": self.variant,
            "face_orders": self.face_orders,
            "edge_orders": self.edge_orders,
        }

    @property
    def lagrange_subdegree(self) -> int:
        poly_coeff = [p.as_sympy().expand().as_poly(*x).as_dict() for p in self._basis]  # type: ignore
        degree = 0
        while True:
            basis_coeff = [
                p.as_sympy().expand().as_poly(*x).as_dict()  # type: ignore
                for p in polynomial_set_1d(self.reference.tdim, degree)
            ]
            monomials = list(set([m for p in poly_coeff + basis_coeff for m in p]))
            mat = sympy.Matrix([[p[m] if m in p else 0 for m in monomials] for p in poly_coeff])
            mat2 = sympy.Matrix(
                [[p[m] if m in p else 0 for m in monomials] for p in poly_coeff + basis_coeff]
            )
            if mat.rank() < mat2.rank():
                return degree - 1
            degree += 1

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return max(
            p.subs(x[2], x[0]).subs(x[1], x[0]).as_sympy().as_poly(x[0]).degree()
            for p in self._basis
        )

    @property
    def polynomial_subdegree(self) -> int:
        return self.lagrange_subdegree

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.lagrange_superdegree

    names = ["transition"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    continuity = "C0"
    value_type = "scalar"
    last_updated = "2023.06"



"""Trimmed serendipity elements on tensor products.

These elements' definitions appear in https://doi.org/10.1137/16M1073352
(Cockburn, Fu, 2017) and https://doi.org/10.1090/mcom/3354
(Gilette, Kloefkorn, 2018)
"""

import typing

from symfem.elements.dpc import DPC, VectorDPC
from symfem.finite_element import CiarletElement
from symfem.functionals import (
    IntegralAgainst,
    IntegralMoment,
    ListOfFunctionals,
    NormalIntegralMoment,
    TangentIntegralMoment,
)
from symfem.functions import FunctionInput, ScalarFunction, VectorFunction
from symfem.moments import make_integral_moment_dofs
from symfem.polynomials import polynomial_set_vector
from symfem.references import NonDefaultReferenceError, Reference
from symfem.symbols import t, x

__all__ = ["TrimmedSerendipityHcurl", "TrimmedSerendipityHdiv"]


class TrimmedSerendipityHcurl(CiarletElement):
    """Trimmed serendipity Hcurl finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order)
        if reference.tdim == 2:
            poly += [
                (x[0] ** j * x[1] ** (order + 1 - j), -(x[0] ** (j + 1)) * x[1] ** (order - j))
                for j in range(order + 1)
            ]
            if order == 0:
                poly += [(x[1], x[0])]
            else:
                poly += [
                    (x[1] ** (order + 1), (order + 1) * x[0] * x[1] ** order),
                    ((order + 1) * x[0] ** order * x[1], x[0] ** (order + 1)),
                ]
        else:
            for i in range(order + 1):
                for j in range(order + 1 - i):
                    for dim in range(3):
                        if i == 0 or dim != 0:
                            p = VectorFunction(tuple(x)).cross(
                                VectorFunction(
                                    [
                                        x[0] ** i * x[1] ** j * x[2] ** (order - i - j)
                                        if d == dim
                                        else 0
                                        for d in range(3)
                                    ]
                                )
                            )
                            poly.append(p)

            if order == 0:
                poly += [ScalarFunction(x[0] * x[1] * x[2] ** (order + 1)).grad(3)]
            else:
                poly += [
                    ScalarFunction(x[0] * x[1] * x[2] ** (order + 1)).grad(3),
                    ScalarFunction(x[0] * x[1] ** (order + 1) * x[2]).grad(3),
                    ScalarFunction(x[0] ** (order + 1) * x[1] * x[2]).grad(3),
                ]
            for i in range(order + 2):
                poly.append(ScalarFunction(x[0] * x[1] ** i * x[2] ** (order + 1 - i)).grad(3))
                if i != 1:
                    poly.append(ScalarFunction(x[1] * x[0] ** i * x[2] ** (order + 1 - i)).grad(3))
                    if order != i:
                        poly.append(
                            ScalarFunction(x[2] * x[0] ** i * x[1] ** (order + 1 - i)).grad(3)
                        )
            for i in range(order + 1):
                p = x[0] * x[1] ** i * x[2] ** (order - i)
                poly.append((0, -x[2] * p, x[1] * p))
                p = x[1] * x[0] ** i * x[2] ** (order - i)
                poly.append((-x[2] * p, 0, x[0] * p))
                if order > 0:
                    p = x[2] * x[0] ** i * x[1] ** (order - i)
                    poly.append((-x[1] * p, x[0] * p, 0))

        dofs: ListOfFunctionals = []
        dofs += make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, DPC, order, {"variant": variant}),
            faces=(IntegralMoment, VectorDPC, order - 2, "contravariant", {"variant": variant}),
            volumes=(IntegralMoment, VectorDPC, order - 4, "contravariant", {"variant": variant}),
        )
        if order >= 1:
            for f_n in range(reference.sub_entity_count(2)):
                for i in range(order + 1):
                    f = ScalarFunction(x[0] ** (order - i) * x[1] ** i).grad(2)
                    f2 = VectorFunction((f[1], -f[0])).subs(x, tuple(t))
                    dofs.append(
                        IntegralAgainst(reference, f2, entity=(2, f_n), mapping="contravariant")
                    )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return (self.order + self.reference.tdim) // (self.reference.tdim + 1)

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order + self.reference.tdim - 1

    names = ["trimmed serendipity Hcurl", "TScurl"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 0
    continuity = "H(curl)"
    value_type = "vector"
    last_updated = "2025.03"


class TrimmedSerendipityHdiv(CiarletElement):
    """Trimmed serendipity Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order)
        if reference.tdim == 2:
            poly += [
                (x[0] ** (j + 1) * x[1] ** (order - j), x[0] ** j * x[1] ** (order + 1 - j))
                for j in range(order + 1)
            ]
            if order == 0:
                poly += [(x[0], -x[1])]
            else:
                poly += [
                    ((order + 1) * x[0] * x[1] ** order, -(x[1] ** (order + 1))),
                    (-(x[0] ** (order + 1)), (order + 1) * x[0] ** order * x[1]),
                ]
        else:
            for i in range(order + 1):
                for j in range(order + 1 - i):
                    p = x[0] ** i * x[1] ** j * x[2] ** (order - i - j)
                    poly.append((x[0] * p, x[1] * p, x[2] * p))
            for i in range(order + 1):
                p = x[0] * x[1] ** i * x[2] ** (order - i)
                poly.append(VectorFunction((0, -x[2] * p, x[1] * p)).curl())
                p = x[1] * x[0] ** i * x[2] ** (order - i)
                poly.append(VectorFunction((-x[2] * p, 0, x[0] * p)).curl())
                if order > 0:
                    p = x[2] * x[0] ** i * x[1] ** (order - i)
                    poly.append(VectorFunction((-x[1] * p, x[0] * p, 0)).curl())

        dofs: ListOfFunctionals = []
        dofs += make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, DPC, order, {"variant": variant}),
            cells=(IntegralMoment, VectorDPC, order - 2, "covariant", {"variant": variant}),
        )
        if order >= 1:
            if reference.tdim == 2:
                fs = [
                    ScalarFunction(x[0] ** (order - i) * x[1] ** i).grad(2)
                    for i in range(order + 1)
                ]
            else:
                fs = [
                    ScalarFunction(x[0] ** (order - i - j) * x[1] ** i * x[2] ** j).grad(3)
                    for i in range(order + 1)
                    for j in range(order + 1 - i)
                ]
            for f in fs:
                f2 = f.subs(x, tuple(t))
                dofs.append(
                    IntegralAgainst(reference, f2, entity=(reference.tdim, 0), mapping="covariant")
                )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return (self.order + 2) // (self.reference.tdim + 1)

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    names = ["trimmed serendipity Hdiv", "TSdiv"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 0
    continuity = "H(div)"
    value_type = "vector"
    last_updated = "2025.03"



"""Orthogonal (Legendre) polynomials."""

import typing

import sympy

from symfem.functions import ScalarFunction
from symfem.symbols import AxisVariablesNotSingle, x

__all__: typing.List[str] = []


def _jrc(
    a, n
) -> typing.Tuple[
    sympy.core.expr.Expr,
    sympy.core.expr.Expr,
    sympy.core.expr.Expr,
]:
    """Get the Jacobi recurrence relation coefficients.

    Args:
        a: The parameter a
        n: The parameter n

    Returns:
        The Jacobi coefficients
    """
    return (
        sympy.Rational((a + 2 * n + 1) * (a + 2 * n + 2), 2 * (n + 1) * (a + n + 1)),
        sympy.Rational(a * a * (a + 2 * n + 1), 2 * (n + 1) * (a + n + 1) * (a + 2 * n)),
        sympy.Rational(n * (a + n) * (a + 2 * n + 2), (n + 1) * (a + n + 1) * (a + 2 * n)),
    )


def orthogonal_basis_interval(
    order: int, derivs: int, variables: AxisVariablesNotSingle = [x[0]]
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 1

    poly = [
        [ScalarFunction(1 if d == 0 else 0) for i in range(order + 1)] for d in range(derivs + 1)
    ]
    for dx in range(derivs + 1):
        for i in range(1, order + 1):
            poly[dx][i] = poly[dx][i - 1] * (2 * variables[0] - 1) * (2 * i - 1) / i
            if dx > 0:
                poly[dx][i] += poly[dx - 1][i - 1] * 2 * dx * (2 * i - 1) / i
            if i > 1:
                poly[dx][i] -= (i - 1) * poly[dx][i - 2] / i
    return poly


def orthogonal_basis_triangle(
    order: int, derivs: int, variables: AxisVariablesNotSingle = [x[0], x[1]]
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 2

    def index(p: int, q: int) -> int:
        """Get the index."""
        return (p + q + 1) * (p + q) // 2 + q

    d_index = index

    poly = [
        [ScalarFunction(0) for i in range((order + 1) * (order + 2) // 2)]
        for d in range((derivs + 1) * (derivs + 2) // 2)
    ]

    for dx in range(derivs + 1):
        for dy in range(derivs + 1 - dx):
            for p in range(order + 1):
                if p == 0:
                    poly[d_index(dx, dy)][index(0, p)] = ScalarFunction(1 if dx == dy == 0 else 0)
                else:
                    pinv = sympy.Rational(1, p)

                    poly[d_index(dx, dy)][index(0, p)] = (
                        poly[d_index(dx, dy)][index(0, p - 1)]
                        * (2 * variables[0] + variables[1] - 1)
                        * (2 - pinv)
                    )
                    if dy > 0:
                        poly[d_index(dx, dy)][index(0, p)] += (
                            poly[d_index(dx, dy - 1)][index(0, p - 1)] * dy * (2 - pinv)
                        )
                    if dx > 0:
                        poly[d_index(dx, dy)][index(0, p)] += (
                            2 * dx * poly[d_index(dx - 1, dy)][index(0, p - 1)] * (2 - pinv)
                        )
                    if p > 1:
                        poly[d_index(dx, dy)][index(0, p)] -= (
                            poly[d_index(dx, dy)][index(0, p - 2)]
                            * (1 - variables[1]) ** 2
                            * (1 - pinv)
                        )
                        if dy > 0:
                            poly[d_index(dx, dy)][index(0, p)] += (
                                2
                                * dy
                                * (1 - variables[1])
                                * poly[d_index(dx, dy - 1)][index(0, p - 2)]
                                * (1 - pinv)
                            )
                        if dy > 1:
                            poly[d_index(dx, dy)][index(0, p)] -= (
                                dy
                                * (dy - 1)
                                * poly[d_index(dx, dy - 2)][index(0, p - 2)]
                                * (1 - pinv)
                            )

                for q in range(1, order - p + 1):
                    a, b, c = _jrc(2 * p + 1, q - 1)
                    poly[d_index(dx, dy)][index(q, p)] = poly[d_index(dx, dy)][index(q - 1, p)] * (
                        (2 * variables[1] - 1) * a + b
                    )
                    if q > 1:
                        poly[d_index(dx, dy)][index(q, p)] -= (
                            poly[d_index(dx, dy)][index(q - 2, p)] * c
                        )
                    if dy > 0:
                        poly[d_index(dx, dy)][index(q, p)] += (
                            poly[d_index(dx, dy - 1)][index(q - 1, p)] * 2 * dy * a
                        )
    return poly


def orthogonal_basis_quadrilateral(
    order: int, derivs: int, variables: AxisVariablesNotSingle = [x[0], x[1]]
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 2

    def d_index(p: int, q: int) -> int:
        """Get the derivative index."""
        return (p + q + 1) * (p + q) // 2 + q

    p0 = orthogonal_basis_interval(order, derivs, [variables[0]])
    p1 = orthogonal_basis_interval(order, derivs, [variables[1]])
    poly = [
        [ScalarFunction(0) for i in range((order + 1) ** 2)]
        for d in range((derivs + 1) * (derivs + 2) // 2)
    ]
    for i in range(derivs + 1):
        for j in range(derivs + 1 - i):
            poly[d_index(i, j)] = [a * b for a in p0[i] for b in p1[j]]
    return poly


def orthogonal_basis_tetrahedron(
    order: int, derivs: int, variables: AxisVariablesNotSingle = x
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 3

    def index(p: int, q: int, r: int) -> int:
        """Get the index."""
        return (p + q + r) * (p + q + r + 1) * (p + q + r + 2) // 6 + (q + r) * (q + r + 1) // 2 + r

    d_index = index

    poly = [
        [ScalarFunction(0) for i in range((order + 1) * (order + 2) * (order + 3) // 6)]
        for d in range((derivs + 1) * (derivs + 2) * (derivs + 3) // 6)
    ]
    for dx in range(derivs + 1):
        for dy in range(derivs + 1 - dx):
            for dz in range(derivs + 1 - dx - dy):
                for p in range(order + 1):
                    if p == 0:
                        poly[d_index(dx, dy, dz)][index(0, 0, p)] = ScalarFunction(
                            1 if dx == dy == dz == 0 else 0
                        )
                    if p > 0:
                        invp = sympy.Rational(1, p)
                        poly[d_index(dx, dy, dz)][index(0, 0, p)] = (
                            poly[d_index(dx, dy, dz)][index(0, 0, p - 1)]
                            * (2 * variables[0] + variables[1] + variables[2] - 1)
                            * (2 - invp)
                        )
                        if dx > 0:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] += (
                                poly[d_index(dx - 1, dy, dz)][index(0, 0, p - 1)]
                                * 2
                                * dx
                                * (2 - invp)
                            )
                        if dy > 0:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] += (
                                poly[d_index(dx, dy - 1, dz)][index(0, 0, p - 1)] * dy * (2 - invp)
                            )
                        if dz > 0:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] += (
                                poly[d_index(dx, dy, dz - 1)][index(0, 0, p - 1)] * dz * (2 - invp)
                            )

                    if p > 1:
                        poly[d_index(dx, dy, dz)][index(0, 0, p)] -= (
                            poly[d_index(dx, dy, dz)][index(0, 0, p - 2)]
                            * (variables[1] + variables[2] - 1) ** 2
                            * (1 - invp)
                        )
                        if dy > 0:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] -= (
                                poly[d_index(dx, dy - 1, dz)][index(0, 0, p - 2)]
                                * 2
                                * (variables[1] + variables[2] - 1)
                                * dy
                                * (1 - invp)
                            )
                        if dy > 1:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] -= (
                                poly[d_index(dx, dy - 2, dz)][index(0, 0, p - 2)]
                                * dy
                                * (dy - 1)
                                * (1 - invp)
                            )
                        if dz > 0:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] -= (
                                poly[d_index(dx, dy, dz - 1)][index(0, 0, p - 2)]
                                * 2
                                * (variables[1] + variables[2] - 1)
                                * dz
                                * (1 - invp)
                            )
                        if dz > 1:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] -= (
                                poly[d_index(dx, dy, dz - 2)][index(0, 0, p - 2)]
                                * dz
                                * (dz - 1)
                                * (1 - invp)
                            )
                        if dz > 0 and dy > 0:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] -= (
                                poly[d_index(dx, dy - 1, dz - 1)][index(0, 0, p - 2)]
                                * 2
                                * dz
                                * dy
                                * (1 - invp)
                            )

                    for q in range(order - p + 1):
                        if q > 0:
                            a, b, c = _jrc(2 * p + 1, q - 1)
                            poly[d_index(dx, dy, dz)][index(0, q, p)] = poly[d_index(dx, dy, dz)][
                                index(0, q - 1, p)
                            ] * ((2 * variables[1] + variables[2] - 1) * a + (1 - variables[2]) * b)
                            if dy > 0:
                                poly[d_index(dx, dy, dz)][index(0, q, p)] += (
                                    poly[d_index(dx, dy - 1, dz)][index(0, q - 1, p)] * 2 * a * dy
                                )
                            if dz > 0:
                                poly[d_index(dx, dy, dz)][index(0, q, p)] += (
                                    poly[d_index(dx, dy, dz - 1)][index(0, q - 1, p)] * (a - b) * dz
                                )
                            if q > 1:
                                poly[d_index(dx, dy, dz)][index(0, q, p)] -= (
                                    poly[d_index(dx, dy, dz)][index(0, q - 2, p)]
                                    * (1 - variables[2]) ** 2
                                    * c
                                )
                                if dz > 0:
                                    poly[d_index(dx, dy, dz)][index(0, q, p)] += (
                                        poly[d_index(dx, dy, dz - 1)][index(0, q - 2, p)]
                                        * 2
                                        * (1 - variables[2])
                                        * dz
                                        * c
                                    )
                                if dz > 1:
                                    poly[d_index(dx, dy, dz)][index(0, q, p)] -= (
                                        poly[d_index(dx, dy, dz - 2)][index(0, q - 2, p)]
                                        * dz
                                        * (dz - 1)
                                        * c
                                    )

                        for r in range(1, order - p - q + 1):
                            a, b, c = _jrc(2 * p + 2 * q + 2, r - 1)
                            poly[d_index(dx, dy, dz)][index(r, q, p)] = poly[d_index(dx, dy, dz)][
                                index(r - 1, q, p)
                            ] * ((variables[2] * 2 - 1) * a + b)
                            if dz > 0:
                                poly[d_index(dx, dy, dz)][index(r, q, p)] += (
                                    poly[d_index(dx, dy, dz - 1)][index(r - 1, q, p)] * 2 * a * dz
                                )
                            if r > 1:
                                poly[d_index(dx, dy, dz)][index(r, q, p)] -= (
                                    poly[d_index(dx, dy, dz)][index(r - 2, q, p)] * c
                                )

    return poly


def orthogonal_basis_hexahedron(
    order: int, derivs: int, variables: AxisVariablesNotSingle = x
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    if variables is None:
        variables = x
    assert len(variables) == 3

    def d_index(p: int, q: int, r: int) -> int:
        """Get the derivative index."""
        return (p + q + r) * (p + q + r + 1) * (p + q + r + 2) // 6 + (q + r) * (q + r + 1) // 2 + r

    p0 = orthogonal_basis_interval(order, derivs, [variables[0]])
    p1 = orthogonal_basis_interval(order, derivs, [variables[1]])
    p2 = orthogonal_basis_interval(order, derivs, [variables[2]])
    poly = [
        [ScalarFunction(0) for i in range((order + 1) ** 3)]
        for d in range((derivs + 1) * (derivs + 2) * (derivs + 3) // 6)
    ]
    for i in range(derivs + 1):
        for j in range(derivs + 1 - i):
            for k in range(derivs + 1 - i - j):
                poly[d_index(i, j, k)] = [a * b * c for a in p0[i] for b in p1[j] for c in p2[k]]
    return poly


def orthogonal_basis_prism(
    order: int, derivs: int, variables: AxisVariablesNotSingle = x
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    if variables is None:
        variables = x
    assert len(variables) == 3

    def d_index_tri(p: int, q: int) -> int:
        """Get the derivative index for a triangle."""
        return (p + q + 1) * (p + q) // 2 + q

    def d_index(p: int, q: int, r: int) -> int:
        """Get the derivative index."""
        return (p + q + r) * (p + q + r + 1) * (p + q + r + 2) // 6 + (q + r) * (q + r + 1) // 2 + r

    p01 = orthogonal_basis_triangle(order, derivs, [variables[0], variables[1]])
    p2 = orthogonal_basis_interval(order, derivs, [variables[2]])
    poly = [
        [ScalarFunction(0) for i in range((order + 1) * (order + 1) * (order + 2) // 2)]
        for d in range((derivs + 1) * (derivs + 2) * (derivs + 3) // 6)
    ]
    for i in range(derivs + 1):
        for j in range(derivs + 1 - i):
            for k in range(derivs + 1 - i - j):
                poly[d_index(i, j, k)] = [a * b for a in p01[d_index_tri(i, j)] for b in p2[k]]
    return poly


def orthogonal_basis_pyramid(
    order: int, derivs: int, variables: AxisVariablesNotSingle = x
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 3

    def index(i: int, j: int, k: int) -> int:
        """Get the index."""
        out = k + j * (order + 1) + i * (order + 1) * (order + 2) // 2 - i * (i**2 + 5) // 6
        if i > j:
            out -= i * (j - 1)
        else:
            out -= j * (j - 1) // 2 + i * (i - 1) // 2
        return out

    def d_index(p: int, q: int, r: int) -> int:
        """Get the derivative index."""
        return (p + q + r) * (p + q + r + 1) * (p + q + r + 2) // 6 + (q + r) * (q + r + 1) // 2 + r

    def combinations(n: int, k: int) -> int:
        """Get the number of combinations."""
        out = 1
        for i in range(n, n - k, -1):
            out *= i
        return out

    poly = [
        [ScalarFunction(0) for i in range((2 * order + 3) * (order + 2) * (order + 1) // 6)]
        for d in range((derivs + 1) * (derivs + 2) * (derivs + 3) // 6)
    ]
    for dx in range(derivs + 1):
        for dy in range(derivs + 1 - dx):
            for dz in range(derivs + 1 - dx - dy):
                poly[d_index(dx, dy, dz)][0] = ScalarFunction(1 if dx == dy == dz == 0 else 0)

                for i in range(order + 1):
                    if i > 0:
                        poly[d_index(dx, dy, dz)][index(i, 0, 0)] = (
                            poly[d_index(dx, dy, dz)][index(i - 1, 0, 0)]
                            * (2 * variables[0] + variables[2] - 1)
                            * (2 * i - 1)
                            / i
                        )
                        if dx > 0:
                            poly[d_index(dx, dy, dz)][index(i, 0, 0)] += (
                                poly[d_index(dx - 1, dy, dz)][index(i - 1, 0, 0)]
                                * 2
                                * dx
                                * (2 * i - 1)
                                / i
                            )
                        if dz > 0:
                            poly[d_index(dx, dy, dz)][index(i, 0, 0)] += (
                                poly[d_index(dx, dy, dz - 1)][index(i - 1, 0, 0)]
                                * dz
                                * (2 * i - 1)
                                / i
                            )
                    if i > 1:
                        poly[d_index(dx, dy, dz)][index(i, 0, 0)] -= (
                            poly[d_index(dx, dy, dz)][index(i - 2, 0, 0)]
                            * (1 - variables[2]) ** 2
                            * (i - 1)
                            / i
                        )
                        if dz > 0:
                            poly[d_index(dx, dy, dz)][index(i, 0, 0)] += (
                                poly[d_index(dx, dy, dz - 1)][index(i - 2, 0, 0)]
                                * 2
                                * (1 - variables[2])
                                * dz
                                * (i - 1)
                                / i
                            )
                        if dz > 1:
                            poly[d_index(dx, dy, dz)][index(i, 0, 0)] -= (
                                poly[d_index(dx, dy, dz - 2)][index(i - 2, 0, 0)]
                                * dz
                                * (dz - 1)
                                * (i - 1)
                                / i
                            )

                    for j in range(order + 1):
                        if j > 0:
                            if i >= j:
                                poly[d_index(dx, dy, dz)][index(i, j, 0)] = (
                                    poly[d_index(dx, dy, dz)][index(i, j - 1, 0)]
                                    * (2 * variables[1] / (1 - variables[2]) - 1)
                                    * (2 * j - 1)
                                    / j
                                )
                                if dy > 0:
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                        poly[d_index(dx, dy - 1, dz)][index(i, j - 1, 0)]
                                        * 2
                                        * dy
                                        / (1 - variables[2])
                                        * (2 * j - 1)
                                        / j
                                    )
                                for di in range(1, dz + 1):
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                        poly[d_index(dx, dy, dz - di)][index(i, j - 1, 0)]
                                        * combinations(dz, di)
                                        * variables[1]
                                        / (1 - variables[2]) ** (di + 1)
                                        * 2
                                        * (2 * j - 1)
                                        / j
                                    )
                                    if dy > 0:
                                        poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                            poly[d_index(dx, dy - 1, dz - di)][index(i, j - 1, 0)]
                                            * combinations(dz, di)
                                            * dy
                                            / (1 - variables[2]) ** (di + 1)
                                            * 2
                                            * (2 * j - 1)
                                            / j
                                        )
                            else:
                                poly[d_index(dx, dy, dz)][index(i, j, 0)] = (
                                    poly[d_index(dx, dy, dz)][index(i, j - 1, 0)]
                                    * (2 * variables[1] + variables[2] - 1)
                                    * (2 * j - 1)
                                    / j
                                )
                                if dy > 0:
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                        poly[d_index(dx, dy - 1, dz)][index(i, j - 1, 0)]
                                        * 2
                                        * dy
                                        * (2 * j - 1)
                                        / j
                                    )
                                if dz > 0:
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                        poly[d_index(dx, dy, dz - 1)][index(i, j - 1, 0)]
                                        * dz
                                        * (2 * j - 1)
                                        / j
                                    )
                        if j > 1:
                            if i >= j:
                                poly[d_index(dx, dy, dz)][index(i, j, 0)] -= (
                                    poly[d_index(dx, dy, dz)][index(i, j - 2, 0)] * (j - 1) / j
                                )
                            elif i + 1 == j:
                                poly[d_index(dx, dy, dz)][index(i, j, 0)] -= (
                                    poly[d_index(dx, dy, dz)][index(i, j - 2, 0)]
                                    * (1 - variables[2])
                                    * (j - 1)
                                    / j
                                )
                                if dz > 0:
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                        poly[d_index(dx, dy, dz - 1)][index(i, j - 2, 0)]
                                        * dz
                                        * (j - 1)
                                        / j
                                    )
                            else:
                                poly[d_index(dx, dy, dz)][index(i, j, 0)] -= (
                                    poly[d_index(dx, dy, dz)][index(i, j - 2, 0)]
                                    * (1 - variables[2]) ** 2
                                    * (j - 1)
                                    / j
                                )
                                if dz > 0:
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                        poly[d_index(dx, dy, dz - 1)][index(i, j - 2, 0)]
                                        * 2
                                        * dz
                                        * (1 - variables[2])
                                        * (j - 1)
                                        / j
                                    )
                                if dz > 1:
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] -= (
                                        poly[d_index(dx, dy, dz - 2)][index(i, j - 2, 0)]
                                        * dz
                                        * (dz - 1)
                                        * (j - 1)
                                        / j
                                    )

                        for k in range(1, order + 1 - max(i, j)):
                            a, b, c = _jrc(2 * max(i, j) + 2, k - 1)
                            poly[d_index(dx, dy, dz)][index(i, j, k)] = (
                                poly[d_index(dx, dy, dz)][index(i, j, k - 1)]
                                * a
                                * (2 * variables[2] - 1)
                            )
                            if dz > 0:
                                poly[d_index(dx, dy, dz)][index(i, j, k)] += (
                                    poly[d_index(dx, dy, dz - 1)][index(i, j, k - 1)] * a * 2 * dz
                                )
                            poly[d_index(dx, dy, dz)][index(i, j, k)] += (
                                b * poly[d_index(dx, dy, dz)][index(i, j, k - 1)]
                            )
                            if k > 1:
                                poly[d_index(dx, dy, dz)][index(i, j, k)] -= (
                                    c * poly[d_index(dx, dy, dz)][index(i, j, k - 2)]
                                )

    return poly


def orthogonal_basis(
    cell: str, order: int, derivs: int, variables: typing.Optional[AxisVariablesNotSingle] = None
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        cell: The cell type
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    args: typing.List[typing.Any] = [order, derivs]
    if variables is not None:
        args.append(variables)

    if cell == "interval":
        return orthogonal_basis_interval(*args)
    if cell == "triangle":
        return orthogonal_basis_triangle(*args)
    if cell == "quadrilateral":
        return orthogonal_basis_quadrilateral(*args)
    if cell == "hexahedron":
        return orthogonal_basis_hexahedron(*args)
    if cell == "tetrahedron":
        return orthogonal_basis_tetrahedron(*args)
    if cell == "prism":
        return orthogonal_basis_prism(*args)
    if cell == "pyramid":
        return orthogonal_basis_pyramid(*args)

    raise ValueError(f"Unsupported cell type: {cell}")


def orthonormal_basis(
    cell: str, order: int, derivs: int, variables: typing.Optional[AxisVariablesNotSingle] = None
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthonormal polynomials.

    Args:
        cell: The cell type
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthonormal polynomials
    """
    from symfem.create import create_reference

    poly = orthogonal_basis(cell, order, derivs, variables)
    ref = create_reference(cell)
    if variables is None:
        variables = x
    norms = [sympy.sqrt((f**2).integral(ref, dummy_vars=variables)) for f in poly[0]]
    for i, n in enumerate(norms):
        for j in range(len(poly)):
            poly[j][i] /= n
    return poly



"""Quadrature definitions."""

import typing

import sympy

__all__ = ["Scalar", "equispaced", "lobatto", "radau", "legendre", "get_quadrature"]

Scalar = typing.Union[sympy.core.expr.Expr, int]


def equispaced(n: int) -> typing.Tuple[typing.List[Scalar], typing.List[Scalar]]:
    """Get equispaced points and weights.

    Args:
        n: Number of points

    Returns:
        Quadrature points and weights
    """
    return (
        [sympy.Rational(i, n - 1) for i in range(n)],
        [
            sympy.Rational(1, 2 * (n - 1)) if i == 0 or i == n - 1 else sympy.Rational(1, n - 1)
            for i in range(n)
        ],
    )


def lobatto(n: int) -> typing.Tuple[typing.List[Scalar], typing.List[Scalar]]:
    """Get Gauss-Lobatto-Legendre points and weights.

    Args:
        n: Number of points

    Returns:
        Quadrature points and weights
    """
    if n == 2:
        return ([0, 1], [sympy.Rational(1, 2), sympy.Rational(1, 2)])
    if n == 3:
        return (
            [0, sympy.Rational(1, 2), 1],
            [sympy.Rational(1, 6), sympy.Rational(2, 3), sympy.Rational(1, 6)],
        )
    if n == 4:
        return (
            [0, (1 - 1 / sympy.sqrt(5)) / 2, (1 + 1 / sympy.sqrt(5)) / 2, 1],
            [
                sympy.Rational(1, 12),
                sympy.Rational(5, 12),
                sympy.Rational(5, 12),
                sympy.Rational(1, 12),
            ],
        )
    if n == 5:
        return (
            [
                0,
                (1 - sympy.sqrt(3) / sympy.sqrt(7)) / 2,
                sympy.Rational(1, 2),
                (1 + sympy.sqrt(3) / sympy.sqrt(7)) / 2,
                1,
            ],
            [
                sympy.Rational(1, 20),
                sympy.Rational(49, 180),
                sympy.Rational(16, 45),
                sympy.Rational(49, 180),
                sympy.Rational(1, 20),
            ],
        )
    if n == 6:
        return (
            [
                0,
                (1 - sympy.sqrt(sympy.Rational(1, 3) + (2 * sympy.sqrt(7) / 21))) / 2,
                (1 - sympy.sqrt(sympy.Rational(1, 3) - (2 * sympy.sqrt(7) / 21))) / 2,
                (1 + sympy.sqrt(sympy.Rational(1, 3) - (2 * sympy.sqrt(7) / 21))) / 2,
                (1 + sympy.sqrt(sympy.Rational(1, 3) + (2 * sympy.sqrt(7) / 21))) / 2,
                1,
            ],
            [
                sympy.Rational(1, 30),
                (14 - sympy.sqrt(7)) / 60,
                (14 + sympy.sqrt(7)) / 60,
                (14 + sympy.sqrt(7)) / 60,
                (14 - sympy.sqrt(7)) / 60,
                sympy.Rational(1, 30),
            ],
        )
    if n == 7:
        return (
            [
                0,
                (1 - sympy.sqrt((5 + 2 * sympy.sqrt(5) / sympy.sqrt(3)) / 11)) / 2,
                (1 - sympy.sqrt((5 - 2 * sympy.sqrt(5) / sympy.sqrt(3)) / 11)) / 2,
                sympy.Rational(1, 2),
                (1 + sympy.sqrt((5 - 2 * sympy.sqrt(5) / sympy.sqrt(3)) / 11)) / 2,
                (1 + sympy.sqrt((5 + 2 * sympy.sqrt(5) / sympy.sqrt(3)) / 11)) / 2,
                1,
            ],
            [
                sympy.Rational(1, 42),
                (124 - 7 * sympy.sqrt(15)) / 700,
                (124 + 7 * sympy.sqrt(15)) / 700,
                sympy.Rational(128, 525),
                (124 + 7 * sympy.sqrt(15)) / 700,
                (124 - 7 * sympy.sqrt(15)) / 700,
                sympy.Rational(1, 42),
            ],
        )
    raise NotImplementedError()


def radau(n: int) -> typing.Tuple[typing.List[Scalar], typing.List[Scalar]]:
    """Get Radau points and weights.

    Args:
        n: Number of points

    Returns:
        Quadrature points and weights
    """
    if n == 2:
        return ([0, sympy.Rational(2, 3)], [sympy.Rational(1, 4), sympy.Rational(3, 4)])
    if n == 3:
        return (
            [0, (6 - sympy.sqrt(6)) / 10, (6 + sympy.sqrt(6)) / 10],
            [sympy.Rational(1, 9), (16 + sympy.sqrt(6)) / 36, (16 - sympy.sqrt(6)) / 36],
        )
    raise NotImplementedError()


def legendre(n: int) -> typing.Tuple[typing.List[Scalar], typing.List[Scalar]]:
    """Get Gauss-Legendre points and weights.

    Args:
        n: Number of points

    Returns:
        Quadrature points and weights
    """
    if n == 1:
        return ([sympy.Rational(1, 2)], [1])
    if n == 2:
        return (
            [(3 - sympy.sqrt(3)) / 6, (3 + sympy.sqrt(3)) / 6],
            [sympy.Rational(1, 2), sympy.Rational(1, 2)],
        )
    if n == 3:
        return (
            [(5 - sympy.sqrt(15)) / 10, sympy.Rational(1, 2), (5 + sympy.sqrt(15)) / 10],
            [sympy.Rational(5, 18), sympy.Rational(4, 9), sympy.Rational(5, 18)],
        )
    raise NotImplementedError()


def get_quadrature(rule: str, n: int) -> typing.Tuple[typing.List[Scalar], typing.List[Scalar]]:
    """Get quadrature points and weights.

    Args:
        rule: The quadrature rule.
              Supported values: equispaced, lobatto, radau, legendre, gll

        n: Number of points

    Returns:
        Quadrature points and weights
    """
    if rule == "equispaced":
        return equispaced(n)
    if rule == "lobatto":
        return lobatto(n)
    if rule == "gll":
        return lobatto(n)
    if rule == "radau":
        return radau(n)
    if rule == "legendre":
        return legendre(n)
    raise ValueError(f"Unknown quadrature rule: {rule}")



"""Regge elements on simplices.

This element's definition appears in https://doi.org/10.1007/BF02733251
(Regge, 1961), https://doi.org/10.1007/s00211-011-0394-z
(Christiansen, 2011), and
http://aurora.asc.tuwien.ac.at/~mneunteu/thesis/doctorthesis_neunteufel.pdf
(Neunteufel, 2021)
"""

import typing
from itertools import product

import sympy

from symfem.elements.lagrange import Lagrange
from symfem.finite_element import CiarletElement
from symfem.functionals import (
    InnerProductIntegralMoment,
    IntegralAgainst,
    IntegralMoment,
    ListOfFunctionals,
    PointInnerProduct,
)
from symfem.functions import FunctionInput
from symfem.moments import make_integral_moment_dofs
from symfem.polynomials import polynomial_set_vector
from symfem.references import Reference
from symfem.symbols import t, x

__all__ = ["Regge", "ReggeTP"]


class Regge(CiarletElement):
    """A Regge element on a simplex."""

    def __init__(self, reference: Reference, order: int, variant: str = "point"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        from symfem import create_reference

        poly: typing.List[FunctionInput] = []
        if reference.tdim == 2:
            poly = [
                ((p[0], p[1]), (p[1], p[2]))
                for p in polynomial_set_vector(reference.tdim, 3, order)
            ]
        if reference.tdim == 3:
            poly = [
                ((p[0], p[1], p[3]), (p[1], p[2], p[4]), (p[3], p[4], p[5]))
                for p in polynomial_set_vector(reference.tdim, 6, order)
            ]

        dofs: ListOfFunctionals = []
        if variant == "point":
            for edim in range(1, 4):
                et = reference.sub_entity_types[edim]
                for e_n in range(reference.sub_entity_count(edim)):
                    assert isinstance(et, str)
                    entity = reference.sub_entity(edim, e_n)
                    for i in product(range(1, order + 2), repeat=edim):
                        if sum(i) < order + 2:
                            for edge in entity.edges[::-1]:
                                tangent = tuple(
                                    b - a
                                    for a, b in zip(
                                        entity.vertices[edge[0]], entity.vertices[edge[1]]
                                    )
                                )
                                dofs.append(
                                    PointInnerProduct(
                                        reference,
                                        tuple(
                                            o
                                            + sum(
                                                sympy.Rational(a[j] * b, order + 2)
                                                for a, b in zip(entity.axes, i[::-1])
                                            )
                                            for j, o in enumerate(entity.origin)
                                        ),
                                        tangent,
                                        tangent,
                                        entity=(edim, e_n),
                                        mapping="double_covariant",
                                    )
                                )

        elif variant == "integral":
            space = Lagrange(create_reference("interval"), order, "equispaced")
            basis = [f.subs(x, t) for f in space.get_basis_functions()]
            for e_n, vs in enumerate(reference.sub_entities(1)):
                edge_e = reference.sub_entity(1, e_n)
                tangent = tuple(
                    (b - a) / edge_e.jacobian()
                    for a, b in zip(edge_e.vertices[0], edge_e.vertices[1])
                )
                for f, dof in zip(basis, space.dofs):
                    dofs.append(
                        InnerProductIntegralMoment(
                            reference,
                            f,
                            tangent,
                            tangent,
                            dof,
                            entity=(1, e_n),
                            mapping="double_covariant",
                        )
                    )

            if reference.tdim == 2:
                if order > 0:
                    dofs += make_integral_moment_dofs(
                        reference,
                        cells=(
                            IntegralMoment,
                            Regge,
                            order - 1,
                            "double_covariant",
                            {"variant": "integral"},
                        ),
                    )

            elif reference.tdim == 3:
                if order > 0:
                    rspace = Regge(create_reference("triangle"), order - 1, "integral")
                    basis = [f.subs(x, t) for f in rspace.get_basis_functions()]
                    for f_n, vs in enumerate(reference.sub_entities(2)):
                        face = reference.sub_entity(2, f_n)
                        for f, dof in zip(basis, rspace.dofs):
                            dofs.append(
                                IntegralMoment(
                                    reference,
                                    f * face.jacobian(),
                                    dof,
                                    entity=(2, f_n),
                                    mapping="double_covariant",
                                )
                            )

                if order > 1:
                    dofs += make_integral_moment_dofs(
                        reference,
                        cells=(
                            IntegralMoment,
                            Regge,
                            order - 2,
                            "double_covariant",
                            {"variant": "integral"},
                        ),
                    )

        else:
            raise ValueError(f"Unknown variant: {variant}")

        super().__init__(
            reference,
            order,
            poly,
            dofs,
            reference.tdim,
            reference.tdim**2,
            (reference.tdim, reference.tdim),
        )
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order

    names = ["Regge"]
    references = ["triangle", "tetrahedron"]
    min_order = 0
    continuity = "inner H(curl)"
    value_type = "symmetric matrix"
    last_updated = "2023.06"


class ReggeTP(CiarletElement):
    """A Regge element on a tensor product cell."""

    def __init__(self, reference: Reference, order: int, variant: str = "integral"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        from symfem import create_reference

        poly: typing.List[FunctionInput] = []
        if reference.tdim == 2:
            for i in range(order + 1):
                for j in range(order + 2):
                    poly.append(((x[0] ** i * x[1] ** j, 0), (0, 0)))
                    poly.append(((0, 0), (0, x[0] ** j * x[1] ** i)))
            for i in range(order + 1):
                for j in range(order + 1):
                    poly.append(((0, x[0] ** i * x[1] ** j), (x[0] ** i * x[1] ** j, 0)))
        elif reference.tdim == 3:
            for i in range(order + 1):
                for j in range(order + 2):
                    for k in range(order + 2):
                        poly.append(
                            ((x[0] ** i * x[1] ** j * x[2] ** k, 0, 0), (0, 0, 0), (0, 0, 0))
                        )
                        poly.append(
                            ((0, 0, 0), (0, x[1] ** i * x[0] ** j * x[2] ** k, 0), (0, 0, 0))
                        )
                        poly.append(
                            ((0, 0, 0), (0, 0, 0), (0, 0, x[2] ** i * x[0] ** j * x[1] ** k))
                        )
            for i in range(order + 1):
                for j in range(order + 1):
                    for k in range(order + 2):
                        poly.append(
                            (
                                (0, x[0] ** i * x[1] ** j * x[2] ** k, 0),
                                (x[0] ** i * x[1] ** j * x[2] ** k, 0, 0),
                                (0, 0, 0),
                            )
                        )
                        poly.append(
                            (
                                (0, 0, x[0] ** i * x[2] ** j * x[1] ** k),
                                (0, 0, 0),
                                (x[0] ** i * x[2] ** j * x[1] ** k, 0, 0),
                            )
                        )
                        poly.append(
                            (
                                (0, 0, 0),
                                (0, 0, x[1] ** i * x[2] ** j * x[0] ** k),
                                (0, x[1] ** i * x[2] ** j * x[0] ** k, 0),
                            )
                        )

        dofs: ListOfFunctionals = []
        if variant == "integral":
            # DOFs on edges
            space = Lagrange(create_reference("interval"), order, "equispaced")
            basis = [f.subs(x, t) for f in space.get_basis_functions()]
            for e_n, vs in enumerate(reference.sub_entities(1)):
                edge = reference.sub_entity(1, e_n)
                tangent = tuple(
                    (b - a) / edge.jacobian() for a, b in zip(edge.vertices[0], edge.vertices[1])
                )
                for f, dof in zip(basis, space.dofs):
                    dofs.append(
                        InnerProductIntegralMoment(
                            reference,
                            f,
                            tangent,
                            tangent,
                            dof,
                            entity=(1, e_n),
                            mapping="double_covariant",
                        )
                    )

            # DOFs on faces
            for f_n, vs in enumerate(reference.sub_entities(2)):
                for i in range(order + 1):
                    for j in range(order + 1):
                        dofs.append(
                            IntegralAgainst(
                                reference,
                                ((0, x[0] ** i * x[1] ** j), (x[0] ** i * x[1] ** j, 0)),
                                entity=(2, f_n),
                                mapping="double_covariant",
                            )
                        )
                for i in range(1, order + 1):
                    for j in range(order + 1):
                        dofs.append(
                            IntegralAgainst(
                                reference,
                                ((x[1] ** i * x[0] ** j * (1 - x[1]), 0), (0, 0)),
                                entity=(2, f_n),
                                mapping="double_covariant",
                            )
                        )
                        dofs.append(
                            IntegralAgainst(
                                reference,
                                ((0, 0), (0, x[0] ** i * x[1] ** j * (1 - x[0]))),
                                entity=(2, f_n),
                                mapping="double_covariant",
                            )
                        )

            if reference.tdim == 3:
                # DOFs on cell
                for i in range(1, order + 1):
                    for j in range(order + 1):
                        for k in range(order + 1):
                            f = x[0] ** i * x[1] ** j * x[2] ** k * (1 - x[0])
                            assert isinstance(f, sympy.core.expr.Expr)
                            dofs.append(
                                IntegralAgainst(
                                    reference,
                                    ((0, 0, 0), (0, 0, f), (0, f, 0)),
                                    entity=(3, 0),
                                    mapping="double_covariant",
                                )
                            )
                            f = x[1] ** i * x[0] ** j * x[2] ** k * (1 - x[1])
                            assert isinstance(f, sympy.core.expr.Expr)
                            dofs.append(
                                IntegralAgainst(
                                    reference,
                                    ((0, 0, f), (0, 0, 0), (f, 0, 0)),
                                    entity=(3, 0),
                                    mapping="double_covariant",
                                )
                            )
                            f = x[2] ** i * x[0] ** j * x[1] ** k * (1 - x[2])
                            assert isinstance(f, sympy.core.expr.Expr)
                            dofs.append(
                                IntegralAgainst(
                                    reference,
                                    ((0, f, 0), (f, 0, 0), (0, 0, 0)),
                                    entity=(3, 0),
                                    mapping="double_covariant",
                                )
                            )
                for i in range(order + 1):
                    for j in range(1, order + 1):
                        for k in range(1, order + 1):
                            f = x[0] ** i * x[1] ** j * x[2] ** k * (1 - x[1]) * (1 - x[2])
                            assert isinstance(f, sympy.core.expr.Expr)
                            dofs.append(
                                IntegralAgainst(
                                    reference,
                                    ((f, 0, 0), (0, 0, 0), (0, 0, 0)),
                                    entity=(3, 0),
                                    mapping="double_covariant",
                                )
                            )
                            f = x[1] ** i * x[0] ** j * x[2] ** k * (1 - x[0]) * (1 - x[2])
                            assert isinstance(f, sympy.core.expr.Expr)
                            dofs.append(
                                IntegralAgainst(
                                    reference,
                                    ((0, 0, 0), (0, f, 0), (0, 0, 0)),
                                    entity=(3, 0),
                                    mapping="double_covariant",
                                )
                            )
                            f = x[2] ** i * x[0] ** j * x[1] ** k * (1 - x[0]) * (1 - x[1])
                            assert isinstance(f, sympy.core.expr.Expr)
                            dofs.append(
                                IntegralAgainst(
                                    reference,
                                    ((0, 0, 0), (0, 0, 0), (0, 0, f)),
                                    entity=(3, 0),
                                    mapping="double_covariant",
                                )
                            )
        else:
            raise ValueError(f"Unknown variant: {variant}")

        super().__init__(
            reference,
            order,
            poly,
            dofs,
            reference.tdim,
            reference.tdim**2,
            (reference.tdim, reference.tdim),
        )
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return (self.order + 1) * self.reference.tdim - 1

    names = ["Regge"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 0
    continuity = "inner H(curl)"
    value_type = "symmetric matrix"
    last_updated = "2023.06"



"""Bubble elements on simplices.

This element's definition appears in https://doi.org/10.1007/978-3-642-23099-8_3
(Kirby, Logg, Rognes, Terrel, 2012)
"""

import typing
from itertools import product

import sympy

from symfem.elements.lagrange import Lagrange
from symfem.finite_element import CiarletElement
from symfem.functionals import DotPointEvaluation, ListOfFunctionals, PointEvaluation
from symfem.functions import FunctionInput
from symfem.references import Reference

__all__ = ["Bubble", "BubbleEnrichedLagrange", "BubbleEnrichedVectorLagrange"]


class Bubble(CiarletElement):
    """Bubble finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        from symfem import create_element

        p1 = create_element(reference.name, "Lagrange", 1, vertices=reference.vertices)
        basis = p1.get_basis_functions()
        bubble = 1
        if reference.name in "quadrilateral":
            bubble *= basis[0]
            bubble *= basis[3]
        elif reference.name in "hexahedron":
            bubble *= basis[0]
            bubble *= basis[7]
        else:
            for f in basis:
                bubble *= f

        if reference.name in ["interval", "quadrilateral", "hexahedron"]:
            o = order - 2
        elif reference.name == "triangle":
            o = order - 3
        elif reference.name == "tetrahedron":
            o = order - 4
        else:
            raise ValueError(f"Unsupported reference: {reference.name}")

        pn = create_element(
            reference.name, "Lagrange", o, vertices=reference.vertices, variant=variant
        )

        poly = [bubble * p for p in pn.get_basis_functions()]

        dofs: ListOfFunctionals = []
        if reference.name in ["interval", "triangle", "tetrahedron"]:

            def func(i):
                return sum(i)
        else:

            def func(i):
                return max(i)

        for i in product(range(1, order), repeat=reference.tdim):
            if func(i) < order:
                point = reference.get_point(tuple(sympy.Rational(j, order) for j in i[::-1]))
                dofs.append(PointEvaluation(reference, point, entity=(reference.tdim, 0)))

        self.variant = variant

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return -1

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order

    @property
    def polynomial_subdegree(self) -> int:
        return -1

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        if self.reference.name in ["quadrilateral", "hexahedron"]:
            return self.order * self.reference.tdim
        return self.order

    names = ["bubble"]
    references = ["interval", "triangle", "tetrahedron", "quadrilateral", "hexahedron"]
    min_order = {
        "interval": 2,
        "triangle": 3,
        "tetrahedron": 4,
        "quadrilateral": 2,
        "hexahedron": 2,
    }
    continuity = "C0"
    value_type = "scalar"
    last_updated = "2024.09"


class BubbleEnrichedLagrange(CiarletElement):
    """Bubble enriched Lagrange element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        lagrange = Lagrange(reference, order, variant)
        bubble = Bubble(reference, order + 2, variant)

        poly: typing.List[FunctionInput] = []
        for e in [lagrange, bubble]:
            for p in e._basis:
                poly.append(p)

        self.variant = variant

        super().__init__(reference, order, poly, lagrange.dofs + bubble.dofs, reference.tdim, 1)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + 2

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order + 2

    names = ["bubble enriched Lagrange"]
    references = ["triangle"]
    min_order = 1
    continuity = "C0"
    value_type = "scalar"
    last_updated = "2024.09"


class BubbleEnrichedVectorLagrange(CiarletElement):
    """Bubble enriched Lagrange element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        lagrange = Lagrange(reference, order, variant)
        bubble = Bubble(reference, order + 2, variant)

        poly: typing.List[FunctionInput] = []
        for e in [lagrange, bubble]:
            for p in e._basis:
                poly.append((p, 0))
                poly.append((0, p))

        dofs: ListOfFunctionals = [
            DotPointEvaluation(reference, d.dof_point(), v, entity=d.entity)
            for d in lagrange.dofs + bubble.dofs
            for v in [(1, 0), (0, 1)]
        ]

        self.variant = variant

        super().__init__(reference, order, poly, dofs, reference.tdim, 2)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + 2

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order + 2

    names = ["bubble enriched vector Lagrange"]
    references = ["triangle"]
    min_order = 1
    continuity = "C0"
    value_type = "vector"
    last_updated = "2024.09"



"""Mardal-Tai-Winther elements on simplices.

This element's definition appears in https://doi.org/10.1137/S0036142901383910
(Mardal, Tai, Winther, 2002)
and https://doi.org/10.1007/s10092-006-0124-6 (Tail, Mardal, 2006)
"""

import typing

from symfem.elements.lagrange import Lagrange
from symfem.elements.nedelec import NedelecFirstKind
from symfem.finite_element import CiarletElement
from symfem.functionals import (
    IntegralMoment,
    ListOfFunctionals,
    NormalIntegralMoment,
    TangentIntegralMoment,
)
from symfem.functions import FunctionInput, VectorFunction
from symfem.moments import make_integral_moment_dofs
from symfem.polynomials import polynomial_set_vector
from symfem.references import NonDefaultReferenceError, Reference
from symfem.symbols import x

__all__ = ["MardalTaiWinther"]


class MardalTaiWinther(CiarletElement):
    """Mardal-Tai-Winther Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, Lagrange, 1, "contravariant", {"variant": variant}),
        )

        poly: typing.List[FunctionInput] = []
        if reference.name == "triangle":
            poly += [
                (1, 0),
                (x[0], 0),
                (x[1], 0),
                (0, 1),
                (0, x[0]),
                (0, x[1]),
                # (x**2 + 2*x*y, -2*x*y - y**2)
                (x[0] ** 2 + 2 * x[0] * x[1], -2 * x[0] * x[1] - x[1] ** 2),
                # (-x**3 + 2*x**2 + 3*x*y**2, 3*x**2*y - 4*x*y - y**3)
                (
                    -(x[0] ** 3) + 2 * x[0] ** 2 + 3 * x[0] * x[1] ** 2,
                    3 * x[0] ** 2 * x[1] - 4 * x[0] * x[1] - x[1] ** 3,
                ),
                # (2*x**2*y + x**2 + 3*x*y**2, -2*x*y**2 - 2*x*y - y**3)
                (
                    2 * x[0] ** 2 * x[1] + x[0] ** 2 + 3 * x[0] * x[1] ** 2,
                    -2 * x[0] * x[1] ** 2 - 2 * x[0] * x[1] - x[1] ** 3,
                ),
            ]
            dofs += make_integral_moment_dofs(
                reference,
                facets=(TangentIntegralMoment, Lagrange, 0, "contravariant", {"variant": variant}),
            )
        else:
            assert reference.name == "tetrahedron"

            poly += polynomial_set_vector(reference.tdim, reference.tdim, 1)
            for p in polynomial_set_vector(reference.tdim, reference.tdim, 1):
                poly.append(
                    VectorFunction(
                        tuple(i * x[0] * x[1] * x[2] * (1 - x[0] - x[1] - x[2]) for i in p)
                    ).curl()
                )

            dofs += make_integral_moment_dofs(
                reference,
                facets=(IntegralMoment, NedelecFirstKind, 0, "contravariant", {"variant": variant}),
            )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return 1

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + self.reference.tdim

    @property
    def polynomial_subdegree(self) -> int:
        return self.lagrange_subdegree

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.lagrange_superdegree

    names = ["Mardal-Tai-Winther", "MTW"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    max_order = 1
    continuity = "H(div)"
    value_type = "vector"
    last_updated = "2025.03"
